// ------------------------------------------------------------------------------------------------------------------
//
//     ===================== 
//    ==== R E U D E M O ====          Start mit SYS 8206
//     =====================
//
//
//   Autor   :  Marco Timm
//   Datum   :  13.03.2008
//   Kontakt :  ber www.forum64.de
//   Username:  marco64
//
//   Eine REU ist zwingend notwendig (mindestens: 128KB Ram) um dies Demo zu starten !!!
//   In VICE laesst sich ein REU emulation ber 'Settings -> Catridge/ IO settings -> REU settings' aktivieren.
//
//
// KURZINFO:
//   Scrolldemo fr grosse Bitmaps mit hilfe der REU. 
//   Mit Joy Port2 laesst sich das Bild bewegen. Feuerknopf aendert die Scrollgeschwindigkeit. 
//   Eine "true drive emulation" der Floppy 1541 ist nicht notwendig.
//
//
// INFO:
//   Dieses Demo nutzt die Fhigkeiten der REU um ein 320*200 Bitmap (8000Bytes + 2000Bytes Farbe)
//   innerhalb von 1.5 Frames zu kopieren (fuer HardScroll).
//
//   Das Bild hat insgesammt eine groesse von 872*600 Pixel (64KB Bitmap + 16KB Farbe).
//   Quelle war das GameBoyColor Spiel ZeldaDX. Damit die roh Bilder besser in das 
//   Multicolorformat konvertiert werden konnten, wurden sie um den Faktor2 vergroessert
//   und die farblich etwas bearbeitet.
//
//   Zum Konvertieren wurde ConGo5 verwendet. ( www.editorix.org/congo/ )
//
//
// COMPILIEREN mit cc65 (www.cc65.org):
//   Anweisung unter win32:     cl65.exe reudemo.c -o reudemo.prg --config own.cfg
//   Demostart:                 sys 8206 
//
//
// FILES:
//   reudemo.prg - Hauptprogramm  (start mit SYS 8206)
//   own.cfg     - MemoryModell fuer reudemo.prg   
//   bigxxx.koa  - 9 Teilbilder
//   zelda.snd   - Sound:  Zelda_Overworld_Final_Vers.sid  von http://www.hvsc.c64.org 
//
//
// Zelda_Overworld_Final_Vers.sid info (mit Sidplay2 ausgelesen):
//   Name:            Zelda Overworld (Final Version)
//   Author:          Sascha Zeidler (Linus)
//   Released:        2004 Sascha Zeidler
//   Load range:      $0FF6-$1CAE
//   Init address:    $0FF6
//   Play address:    $1003
//   Besonderheit:    Play muss 3x pro Frame aufgerufen werden
//
//
// MEMORYMAP: (RAM)
//   $0FF6 - $1FFF:   SOUND
//   $2000 - $7FFF:   Dieses C Programm
//   $8000 - $BFFF:   Framebuffer 1
//   $C000 - $FFFF:   Framebuffer 2
//
// ------------------------------------------------------------------------------------------------------------------



#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <peekpoke.h>
#include <c64.h>
#include <6502.h>
#include <cbm.h>


// ---------------------------------------
// ----  Definitionen                  ---
// ---------------------------------------

// Sound
#define INIT_MUSIC	asm("LDA #$00");asm("LDX #$00");asm("LDY #$00");asm("jsr $0FF6");
#define PLAY_MUSIC  asm("LDA #$00");asm("LDX #$00");asm("LDY #$00");asm("jsr $1003");


// Auf Rasterline warten (bezieht sich auf das MSB der Rasterline)
#define WAIT_WHILE_RASTERLINE_LOW    while (!(VIC.ctrl1 & 0x80)) {};
#define WAIT_WHILE_RASTERLINE_HIGH   while (VIC.ctrl1 & 0x80) {};


// VIC Bnke umschalten
#define SET_VIC_BANK1                POKE( 0xDD00, (PEEK(0xDD00) & 0xfc) | 0x11 );
#define SET_VIC_BANK2                POKE( 0xDD00, (PEEK(0xDD00) & 0xfc) | 0x10 );
#define SET_VIC_BANK3                POKE( 0xDD00, (PEEK(0xDD00) & 0xfc) | 0x01 );
#define SET_VIC_BANK4                POKE( 0xDD00, (PEEK(0xDD00) & 0xfc) | 0x00 );


// REU-Register
#define REU_STATUS          0xDF00
#define REU_COMMAND         0xDF01
#define REU_ADR_C64         0xDF02
#define REU_ADR_REU         0xDF04
#define REU_BANK            0xDF06
#define REU_BYTES           0xDF07
#define REU_INTERRUPT       0xDF09   
#define REU_ADRESS_CONTROL  0xDF0A


// AUTOLOAD + AKTIONSTART (stash c64->reu, fetch reu->c64)
#define STASH               252
#define FETCH               253


// REU_xxx(word, word, word, byte);       LESEN = stash;  SCHREIBEN = fetsh
#define REU_LESEN(von,nach,anz,bank)      asm("SEI");POKE(0xDF06,bank);POKEW(0xDF02,von);POKEW(0xDF04,nach);POKEW(0xDF07,anz);POKE(0xDF01,252);asm("CLI");              
#define REU_SCHREIBEN(von,nach,anz,bank)  POKE(0xDF06,bank);POKEW(0xDF04,von);POKEW(0xDF02,nach);POKEW(0xDF07,anz);POKE(0xDF01,253);              


// Bild Formate
//  koala-Format (.koa)
#define LOAD_BITMAP 0000
#define LOAD_COLOR1 8000
#define LOAD_COLOR2 9000
#define LOAD_BACKGROUNDCOLOR 10000

// oxp-Format (.oxp)
//#define LOAD_BITMAP 2048
//#define LOAD_COLOR1 1024
//#define LOAD_COLOR2 0000
//#define LOAD_BACKGROUNDCOLOR 10049


// Multicolorbitmap in groesse 8*8Pixel
#define max_x 109
#define max_y 75


// Joystick definitionen
#define JOY2     0xDC00
#define JOYUP    0x01
#define JOYDOWN  0x02
#define JOYLEFT  0x04
#define JOYRIGHT 0x08
#define JOYFIRE  0x10



// ---------------------------------------
// ----  globale Variablen und Felder  ---
// ---------------------------------------


// variablen die von halfpic1() an halfpic2 uebergeben werden
static char copy_zeile;                        // aktuelle BildZeile
static unsigned int cpupic;                    // c64 adressen, Bitmap 	
static unsigned int cpucol1;                   //  Farbe1
static unsigned int cpucol2;                   //  Farbe2 
static unsigned int reupic;                    // reu adressen
static unsigned int reucol1;
static unsigned int reucol2;

// Spritedaten
static const unsigned char sprite_kleine_box[] = {
	255,255,240,170,170,176,213,85,80,170,170,176,213,85,80,170,170,176,
	213,85,80,170,170,176,213,85,80,170,170,176,213,85,80,255,255,240,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

static const unsigned char sprite_ecke[] = {
	255,255,255,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,
	0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1,0,0,1
};



// ---------------------------------------
// ----  Funktionen                    ---
// ---------------------------------------


// auf REU testen
void testreu() {
	char flag = 1;
	printf("\n Teste auf REU. \n\n");	
	
	// auf REU testen
	while(1) {
		POKEW(0xC000, 0x1234);              // Testdaten  
		REU_LESEN(0xC000,0x0000,2,1);       // 2 Bytes in REU lesen
		POKEW(0xC000, 0x0000);              // Testdaten loeschen 
		SEI();
		REU_SCHREIBEN(0x0000,0xC000,2,1);   // 2 Bytes von REU nach c64 schreiben
		CLI();
		
		if (PEEKW(0xC000) == 0x1234) {        // Testdaten pruefen
			printf(" >>> REU mit mind. 128KB gefunden. <<<\n");
			printf(" > Steuerung : Joystick PORT2        <\n");
			printf(" > Feuerknopf: Geschwindigkeit       <\n\n");
			return;
		}

		if (flag){
			printf("\n Keine REU mit mind. 128KB gefunden.\n");
			printf(" Ein REU ist zwingend notwendig!\n");
			printf(" In VICE REUemu aktivieren ueber:\n");
			printf("  'Settings -> Catridge/ IO settings\n"); 
			printf("  -> REU settings'\n\n");
			flag=0;
		}
	}
}






// Bild von Adresse 'from_adr' mit 'yanz' Zeilen und 'xanz' Spalten in die REU an Position(x,y) einlesen  
static void readpic(unsigned int from_adr, char xpos, char ypos, char xanz, char yanz) {
	
	// lokale variablen
	char i=0;			                    // zaehler

	// globale variablen
	cpupic  = from_adr + LOAD_BITMAP;	    // c64 adressen
	cpucol1 = from_adr + LOAD_COLOR1;
	cpucol2 = from_adr + LOAD_COLOR2;

	reupic  = ypos*max_x*8 + xpos*8;        // reu adressen
	reucol1 = 0x0000 + ypos*max_x + xpos;
	reucol2 = 0x8000 + ypos*max_x + xpos;


	// Bild Zeilen einlesen
	while(i<yanz){        
		REU_LESEN( cpupic, reupic, 8*xanz, 0);         //Bitmap
		cpupic += 40*8;
		reupic += max_x*8;
	
		REU_LESEN( cpucol1, reucol1, xanz, 1);         //color1
		cpucol1 += 40;
		reucol1 += max_x;

		REU_LESEN( cpucol2, reucol2, xanz, 1);         //color2
		cpucol2 += 40;
		reucol2 += max_x;

		i++;
	}
	return;
}


// 17 von 25 Bildzeilen von Position(x,y) in REU nach Framebuffer 'baseadr' kopieren
// Zeitverbrauch: ca. 1Frame
static void halfpic1(char xpos, char ypos, unsigned int baseadr) {
	
	// lokale variablen
	char i=0;

	// globale variablen
	cpupic  = baseadr + 0x2000;             //c64 adressen
	cpucol1 = baseadr;
	cpucol2 = 0xD800;
		
	reupic =  8*max_x*ypos + xpos*8;        //reu adressen
	reucol1 = ypos*max_x + xpos;
	reucol2 = 0x8000 + ypos*max_x + xpos;

	copy_zeile=0;
	
	// 17 von 25 Zeilen in Framebuffer kopieren
	while(copy_zeile<17){
		REU_SCHREIBEN(reupic, cpupic, 320, 0);        // Bitmapzeile kopieren
		cpupic += 8*40;
		reupic += max_x*8;

		REU_SCHREIBEN(reucol1, cpucol1, 40, 1);       // Farbe1 zeile kopieren
		cpucol1 += 40;
		reucol1 += max_x;

		REU_SCHREIBEN(reucol2, cpucol2, 40, 1);       // Farbe2 zeile kopieren
		cpucol2 += 40;
		reucol2 += max_x;

		if (copy_zeile == 6 || copy_zeile == 11) {	 //weile zelda.sid 3x pro Frame aufgerufen werden muss
			PLAY_MUSIC;
		}

		copy_zeile++;
	}
	return;
}



// restliche 8 Bildzeilen von Position(x,y) in REU nach Framebuffer 'baseadr' kopieren
// Zeitverbrauch: ca. 0.5 Frames
static void halfpic2() {

	// Bildzeilen kopieren
	while(copy_zeile<25){
		REU_SCHREIBEN(reupic, cpupic, 320, 0);          // Bitmapzeile kopieren
		cpupic += 8*40;
		reupic += max_x*8;

		REU_SCHREIBEN(reucol1, cpucol1, 40, 1);         // Farbe1 zeile kopieren
		cpucol1 += 40;
		reucol1 += max_x;

		REU_SCHREIBEN(reucol2, cpucol2, 40, 1);         // Farbe2 zeile kopieren
		cpucol2 += 40;
		reucol2 += max_x;

		copy_zeile++;
	}
	copy_zeile=0;
	return;
}


// komplettes Bild kopieren (Hardscroll) und anzeigen (Frame1 aktuelles Bild, Frame2 kopiertes Bild)
// Zeitverbrauch: ca 1.5 Frames
// Info: nach dem 1. Frame wird auf das noch unfertige 2.Frame geblendet das waehrend es angezeigt wird zuendekopiert wird
//       Der Rasterstrahl bleibt immer oberhalb der aktuell kopierten Zeile.
// - Bild von Position(x,y) der Reu kopieren
// - sx , sy  = SCROLL_X und SCROLL_Y werte in frame1 
// - sx2, sy2 = SCROLL_X und SCROLL_Y werte in frame2 
static void paintfullpic(char x, char y, char sx, char sy, char sx2, char sy2) {
	
	static unsigned int paintbase = 0xC000;             //Framebuffer adresse  (statisch weil sie nur einmal initialisiert wird)

	// warten bis neues Frame beginnt
	WAIT_WHILE_RASTERLINE_LOW;
	PLAY_MUSIC;
	WAIT_WHILE_RASTERLINE_HIGH;
	
	// --- Zustand: beginn Frame1 ---
		
	// Scroll register fuer Frame1 setzen
	VIC.ctrl2 = (VIC.ctrl2 & 0xF0) | sx;         // scroll x
	VIC.ctrl1 = (VIC.ctrl1 & 0xF0) | sy;         // scroll y
	PLAY_MUSIC;

	// 1.Bildteil kopieren
	halfpic1(x,y,paintbase);          //Zeitverbrauch: 1 frame


	// --- Zustand: beginn Frame2 ---

	// Darstellung auf 2.Framebuffer umschalten
	if (paintbase == 0xC000){SET_VIC_BANK4;} else {SET_VIC_BANK3;}

	// Scroll register fuer Frame2 setzen
	VIC.ctrl2 = (VIC.ctrl2 & 0xF0) | sx2;         // scroll x
	VIC.ctrl1 = (VIC.ctrl1 & 0xF0) | sy2;         // scroll y

	PLAY_MUSIC;

	// 2.Bildteil kopieren
	halfpic2();


	// --- Zustand: mitte Frame2 ---

	PLAY_MUSIC;

	// fr naechsten Funktionsaufruf auf den nicht sichtbaren framebuffer umstellen
	if (paintbase == 0xC000){ paintbase=0x8000;} else{ paintbase=0xC000;}

	return;
}


// fuer Softscroll
// Zeitverbrauch: ca. 0.5 Frames
// - sx , sy  = SCROLL_X und SCROLL_Y werte fur das neue Frame 
static void paintnopic(char sx, char sy) {

	// warten bis neues Frame beginnt
	WAIT_WHILE_RASTERLINE_LOW;
	PLAY_MUSIC;
	WAIT_WHILE_RASTERLINE_HIGH;
	
	// Scroll register setzen
	VIC.ctrl2 = (VIC.ctrl2 & 0xF0) | sx;         // scroll x
	VIC.ctrl1 = (VIC.ctrl1 & 0xF0) | sy;         // scroll y

	PLAY_MUSIC;
	
	// warten bis Raster ca. 1/3 des Bildschirms erreicht hat
	while (VIC.rasterline < 130) {};
	
	PLAY_MUSIC;

	return;
}


// keine veraenderung am bild, aber die Music laeuft weiter
// Zeitverbrauch: ca. 0.5 Frames
static void waitpic() {

	// warten bis neues Frame beginnt	
	WAIT_WHILE_RASTERLINE_LOW;
	PLAY_MUSIC;
	WAIT_WHILE_RASTERLINE_HIGH;
	
	PLAY_MUSIC;

	// warten bis Raster ca. 1/3 des Bildschirms erreicht hat
	while (VIC.rasterline < 130) {};
	
	PLAY_MUSIC;
	return;
}


// ---------------------------------------
// ----  Hauptprogramm                 ---
// ---------------------------------------

int main (void)
{
	// lokale Variablen
	char sx,sy;                        // Softscroll
	signed char richtung_x = 0;        // Richtungsflag X
	signed char richtung_y = 0;        // Richtungsflag Y
	char fire=0;                       // Joystick Feuerknopf Flag
	char x=0;                          // Bildpos x in der REU
	char y=0;                          // Bildpos y in der REU
	char joy;                          // Joystickwert zwischenspeicher
	char i=0;                          // zahler
	
	// --- Initialisierungen ---

	testreu();
	printf("\nBitte warten,lade BILDER und SOUND nach.");

	// Bilder laden und in REU kopieren
	cbm_load("big1x1.koa", 8, (void*) 0x8000);
	readpic(0x8000, 0, 0,40,25);
	cbm_load("big1x2.koa", 8, (void*) 0x8000);
	readpic(0x8000, 0,25,40,25);
	cbm_load("big1x3.koa", 8, (void*) 0x8000);
	readpic(0x8000, 0,50,40,25);

	cbm_load("big2x1.koa", 8, (void*) 0x8000);
	readpic(0x8000, 40, 0,40,25);
	cbm_load("big2x2.koa", 8, (void*) 0x8000);
	readpic(0x8000, 40,25,40,25);
	cbm_load("big2x3.koa", 8, (void*) 0x8000);
	readpic(0x8000, 40,50,40,25);

	cbm_load("big3x1.koa", 8, (void*) 0x8000);
	readpic(0x8000, 80, 0,29,25);
	cbm_load("big3x2.koa", 8, (void*) 0x8000);
	readpic(0x8000, 80,25,29,25);
	cbm_load("big3x3.koa", 8, (void*) 0x8000);
	readpic(0x8000, 80,50,29,25);

	
	// Hintergrund Farbe setzen
	VIC.bgcolor0 = 0;

	// Musik laden
	cbm_load("zelda.snd", 8, (void*) 0x0FF6);	
	INIT_MUSIC;

	// VIC register setzen
	VIC.ctrl1 = (VIC.ctrl1 & 0xf0) | 0x28;  //Bitmap an 
	VIC.ctrl2 = (VIC.ctrl2 & 0xf0) | 0x18;  //Multicolor an
	VIC.addr  = 0x08;                       //Bitmap an $2000(VIC)  
	SET_VIC_BANK4;
  

	// Sprites kopieren und initialisieren
	for(i=0; i<63; i++) {                // Sprite daten in VIC Bank 3 und 4 kopieren
		POKE(0x8000 + 0x0400 + i, sprite_kleine_box[i]);
		POKE(0xc000 + 0x0400 + i, sprite_kleine_box[i]);
		POKE(0x8000 + 0x0400 + 64 + i, sprite_ecke[i]);
		POKE(0xc000 + 0x0400 + 64 + i, sprite_ecke[i]);
	};
	POKE(0x8000 + 0x03f8 , 16); POKE(0x8000 + 0x03f9 , 17);      // Spritezeiger in Bank3 setzen
	POKE(0xc000 + 0x03f8 , 16); POKE(0xc000 + 0x03f9 , 17);      // Spritezeiger in Bank4 setzen
	
	VIC.spr_ena = 1 | 2;                                // Sprites einschalten
	VIC.spr0_x = 100; VIC.spr0_y = 100;                 // Sprite 1 setzen  
	VIC.spr1_x = 31+55-24+2;  VIC.spr1_y = 207 -1;      // Sprite 2 setzen
	VIC.spr0_color = 1; VIC.spr1_color = 1;             // Spritefarben setzen
	

	// Bild zeichnen
	paintfullpic(x,y,4,4,4,4);
	
	
	// Auf richtigen Zeitpunkt warten ( 
	WAIT_WHILE_RASTERLINE_LOW;
	SEI();     // Interrupt verbieten
	
	
	//Programm schleife (Endlosschleife)
	while(1){  


		if ( richtung_x || richtung_y ) {     // wenn Bewegung
			if (!fire) {                          // wenn kein Feuerknopf (4Pixel Scrolling pro Frame) 
				sx=4; sy=4;
				if (richtung_x== 1) sx=0;         // Scrollregister werte fr 1.Frame festlegen (siehe paintfullpic()) 
				if (richtung_x==-1) sx=7;
				if (richtung_y== 1) sy=0;
				if (richtung_y==-1) sy=7;
				paintfullpic(x,y,sx,sy,4,4);      // Scrollen
			} else {							    // wenn Feuerknopf (1Pixel Scrolling pro Frame)
				sx=4; sy=4;
				if (richtung_x != 0) { y -= richtung_y; richtung_y=0; }    // nur X oder Y Scrolling zulassen
				
				i=0;
				while (i<7) {                         // 8Pixel Scrollen (soft+hard)
					sx -= richtung_x;
					sy -= richtung_y;
					if (sx==0 && richtung_x==1) {           // Hardscroll nach rechts
						sx=7;
						paintfullpic(x, y, 0, sy, 7, sy);
					} else if (sx==7 && richtung_x==-1) {   // Hardscroll nach links
						sx=0;
						paintfullpic(x, y, 7, sy, 0, sy);
					} else if (sy==0 && richtung_y==1) {    // Hardscroll nach unten
						sy=7;
						paintfullpic(x, y, sx , 0, sx, 7);
					} else if (sy==7 && richtung_y==-1) {   // Hardscroll nach oben
						sy=0;
						paintfullpic(x, y, sx,  7, sx, 0);
					} else {                                // Softscroll
						paintnopic(sx, sy);    
					}
					i++;
				}
			}
		} else {          // sonst 1Frame warten
			waitpic(); 
		}


		// Joystickabfrage
		joy=PEEK(JOY2) & 0x1f;
		richtung_x=0;
		richtung_y=0;
		fire=0;
		if (!(joy & JOYFIRE))   fire = 1; 
		if (!(joy & JOYLEFT))   richtung_x = -1; 
		if (!(joy & JOYRIGHT))  richtung_x =  1;
		if (!(joy & JOYUP))     richtung_y = -1; 
		if (!(joy & JOYDOWN))   richtung_y =  1;
		
		
		// (x,y) position des Bild in der REU neu setzen und Grenzen abfragen
		x += richtung_x;
		y += richtung_y;
		
		if (x==255)   {x=0; richtung_x=0;}
		if (x==max_x+1-40) {x=max_x-40; richtung_x=0;}
		if (y==255)   {y=0; richtung_y=0;}
		if (y==max_y+1-25) {y=max_y-25; richtung_y=0;}


		// Spritebox setzen
		VIC.spr0_x = x/2 + 31 +1;
		VIC.spr0_y = y/2 + 246 - 37 -1;
	}


	return EXIT_SUCCESS;
}


